import sys
import StringIO
from struct import pack, calcsize
from PIL import Image
import argparse
from os.path import os

class OBJ():
  def __init__(self):
    self.vs = []
    self.ns = []
    self.ts = []
    self.fs = []

def readOBJ(str):
  obj = OBJ()
  def parseVertex(coords):
    obj.vs.append(tuple([float(c) for c in coords]))
  def parseTex(coords):
    obj.ts.append(tuple([float(c) for c in coords]))
  def parseNormal(coords):
    obj.ns.append(tuple([float(c) for c in coords]))
  def parseFace(indicies):
    obj.fs.append(tuple(tuple(int(i) for i in index.split("/")) for index in indicies))
  f = StringIO.StringIO(str)
  parse = { "v" : parseVertex,
            "vt" : parseTex,
            "vn" : parseNormal,
            "f" : parseFace
        }
  for line in f:
    line = line.strip()
    if line.startswith("#"):
      continue
    parts = line.split(" ")
    if not parts[0] in parse:
      continue
    parse[parts[0]](parts[1:])
  return obj

def obj2mf3d(obj, file, fmt, tex=("none", None), mask=("none", None)):
  def pack_data(obj, fmt):
    def _v(v, n, t):
      return v
    def _vn(v, n, t):
      return v + n
    def _vnt(v, n, t):
      return v + n + t
    def _vt(v, n, t):
      return v + t
    fmts = {"v3f" : ("<3f", _v),
          "v3f_n3f" : ("<3f3f", _vn),
          "v3f_n3f_t2f" : ("<3f3f2f", _vnt),
          "v3f_t2f" : ("<3f2f", _vt)}
    vertex, func = fmts[fmt]
    data = "".join(pack(vertex, *func(v, n, t)) for v, n, t in zip(obj.vs, obj.ns, obj.ts))
    assert len(data) == calcsize(vertex) * len(obj.vs)
    return data
    
  fmts = {"v3f" : 0,
          "v3f_n3f" : 1,
          "v3f_n3f_t2f" : 2,
          "v3f_t2f" : 3}
  tex_fmts = {"none" : 0,
              "jpg" : 1,
              "png" : 2}
  vertex_data = pack_data(obj, fmt)
  index_data = pack("<%dH" % len(obj.fs), *obj.fs)
  
  file.write(pack("<4c", *"MF3D")) # header
  file.write(pack("<BBB", fmts[fmt], tex_fmts[tex[0]], tex_fmts[mask[0]])) # data, texture, mask fmt
  file.write(pack("<IIII",  len(vertex_data),
                            len(index_data),
                            len(tex[1]) if tex[1] else 0, 
                            len(mask[1]) if mask[1] else 0)) # data, index, texture, mask size
  file.write(vertex_data)
  file.write(index_data)
  if tex[1]:
    file.write(tex[1])
  if mask[1]:
    file.write(mask[1])
  file.flush()

def reindexOBJ(obj):
  robj = OBJ()
  i = 0
  fs = {}
  for ts in obj.fs:
    for t in ts:
      if t not in fs:
        fs[t] = i
        robj.vs.append(obj.vs[t[0] - 1])
        robj.ts.append(obj.ts[t[1] - 1])
        robj.ns.append(obj.ns[t[2] - 1])
        i += 1
      robj.fs.append(fs[t])
  #assert len(obj.vs) == len(robj.vs)
  #assert len(obj.ts) == len(robj.ts)
  #assert len(obj.ns) == len(robj.ns)
  return robj

def make_mask(data):
  img = Image.open(StringIO.StringIO(data))
  mask = Image.new("1", img.size)
  mpixels = mask.load()
  pixels = img.load()
  for x in range(img.size[0]):
    for y in range(img.size[1]):
      r, g, b = pixels[x, y]
      mpixels[x, y] = 255 if (r > 150 and g < 50 and b > 150) else 0
  out = StringIO.StringIO()
  mask.save(out, "png", optimize=True)
  return out.getvalue()

def export(src, dst, fmt, tex=("none", None), mask=("none", None)):
  obj = reindexOBJ(readOBJ(src.read()))
  obj2mf3d(obj, dst, fmt, tex, mask)
  dst.close()

if __name__=="__main__":
  fmt = {"v" : "v3f", "vn" : "v3f_n3f", "vt" : "v3f_t2f" , "vnt" : "v3f_n3f_t2f" }
  p = argparse.ArgumentParser(description="Convert .OBJ model -> .MF3D")
  p.add_argument("obj", type=file)
  p.add_argument("mf3d", type=argparse.FileType("wb"))
  p.add_argument("-t", "--texture", required=False)
  p.add_argument("-vfmt", "--vertex_format", help="vertex format", required=True, choices=fmt.keys())
  p.add_argument("-m", "--mask", action="store_true", help="generate mask")
  
  if len(sys.argv) == 1:
    p.print_help()
    sys.exit(1)
  args = p.parse_args()
  tex = ("none", None)
  if args.texture:
    ext = os.path.splitext(args.texture)[1][1:]
    tex = (ext, open(args.texture, "rb").read())
  export(args.obj, args.mf3d, fmt[args.vertex_format], tex)
